@@ -1,5 +1,5 @@
-/* 
- *  at.c : Put file into atrun queue
+/*
+ *  at.c : Put file into atd queue
  *  Copyright (C) 1993, 1994, 1995, 1996, 1997 Thomas Koenig
  *  Copyright (C) 2002, 2005 Ryan Murray
  *
@@ -78,7 +78,7 @@
 #include "panic.h"
 #include "parsetime.h"
 #include "perm.h"
-
+#include "posixtm.h"
 #include "privs.h"
 
 /* Macros */
@@ -94,6 +94,9 @@
 #define TIMEFORMAT_POSIX	"%a %b %e %T %Y"
 #define TIMESIZE	50
 
+#define DEFAULT_QUEUE 'a'
+#define BATCH_QUEUE   'b'
+
 enum {
     ATQ, BATCH, ATRM, AT, CAT
 };				/* what program we want to run */
@@ -108,7 +111,6 @@ gid_t daemon_gid = (gid_t) - 3;
 
 /* File scope variables */
 
-static const char *svnid = "$Id$";
 char *no_export[] =
 {
     "TERM", "DISPLAY", "_", "SHELLOPTS", "BASH_VERSINFO", "EUID", "GROUPS", "PPID", "UID"
@@ -125,6 +127,7 @@ char atfile[] = ATJOB_DIR "/12345678901234";
 char *atinput = (char *) 0;	/* where to get input from */
 char atqueue = 0;		/* which queue to examine for jobs (atq) */
 char atverify = 0;		/* verify time instead of queuing job */
+char *mail_rcpt = (char *) 0;   /* user to send mail to */
 
 /* Function declarations */
 
@@ -133,7 +136,6 @@ static void alarmc(int signo);
 static char *cwdname(void);
 static void writefile(time_t runtimer, char queue);
 static void list_jobs(void);
-static time_t parsetimespec(const char *spec);
 
 /* Signal catching functions */
 
@@ -209,7 +211,7 @@ nextjob()
 	fscanf(fid, "%5lx", &jobno);
 	rewind(fid);
     } else {
-	fid = fopen(ATJOB_DIR "/.SEQ", "w");
+	fid = fopen(LFILE, "w");
 	if (fid == NULL)
 	    return EOF;
     }
@@ -351,22 +353,36 @@ writefile(time_t runtimer, char queue)
     if ((fp = fdopen(fd, "w")) == NULL)
 	panic("Cannot reopen atjob file");
 
-    /* Get the userid to mail to, first by trying getlogin(), which reads
-     * /var/run/utmp, then from LOGNAME, finally from getpwuid().
-     */
-    mailname = getlogin();
-    if (mailname == NULL)
-	mailname = getenv("LOGNAME");
-    if (mailname == NULL || mailname[0] == '\0' || getpwnam(mailname) == NULL) {
-	pass_entry = getpwuid(real_uid);
-	if (pass_entry != NULL)
-	    mailname = pass_entry->pw_name;
+    if (mail_rcpt != NULL)
+        /* If the userid to mail to has been set on the command-line, then
+         * validate the user and continue
+         */
+        if (getpwnam(mail_rcpt) != NULL)
+            mailname = mail_rcpt;
+        else {
+            char msg[80];
+            snprintf(msg, sizeof(msg), "Cannot find username %s", mail_rcpt);
+            panic(msg);
+        }
+    else {
+        /* Get the userid to mail to, first by trying getlogin(), which reads
+         * /var/run/utmp, then from LOGNAME, finally from getpwuid().
+         */
+        mailname = getlogin();
+        if (mailname == NULL)
+            mailname = getenv("LOGNAME");
+        if (mailname == NULL || mailname[0] == '\0' || getpwnam(mailname) == NULL) {
+            pass_entry = getpwuid(real_uid);
+            if (pass_entry != NULL)
+	        mailname = pass_entry->pw_name;
+        }
     }
 
     if ((mailname == NULL) || (mailname[0] == '\0')
 	|| (strlen(mailname) > mailsize) ) {
 	panic("Cannot find username to mail output to");
     }
+
     if (atinput != (char *) NULL) {
 	fpin = freopen(atinput, "r", stdin);
 	if (fpin == NULL)
@@ -389,6 +405,22 @@ writefile(time_t runtimer, char queue)
 	int export = 1;
 	char *eqp;
 
+        /* Only accept alphanumerics and underscore in variable names.
+         * Also require the name to not start with a digit.
+         * Some shells don't like other variable names.
+         */
+        {
+            char *p = *atenv;
+            if (isdigit(*p))
+                export = 0;
+            for (; *p != '=' && *p != '\0'; ++p) {
+                if (!isalnum(*p) && *p != '_') {
+                    export = 0;
+                    break;
+                }
+            }
+        }
+
 	eqp = strchr(*atenv, '=');
 	if (ap == NULL)
 	    eqp = *atenv;
@@ -462,6 +494,11 @@ writefile(time_t runtimer, char queue)
 
     istty = isatty(fileno(stdin));
     if (istty) {
+	runtime = localtime(&runtimer);
+
+	strftime(timestr, TIMESIZE, TIMEFORMAT_POSIX, runtime);
+	fprintf(stderr, "at %s\n", timestr);
+
 	fprintf(stderr, "at> ");
 	fflush(stderr);
     }
@@ -476,6 +513,9 @@ writefile(time_t runtimer, char queue)
 	fprintf(stderr, "<EOT>\n");
     }
     fprintf(fp, "\n");
+    if (ferror(fp))
+	panic("Output error");
+    fflush(fp);
     if (ferror(fp))
 	panic("Output error");
 
@@ -492,6 +532,7 @@ writefile(time_t runtimer, char queue)
 
     close(fd2);
 
+    /* This line maybe superfluous after commit 11cb731bb560eb7bff4889c5528d5f776606b0d3 */
     runtime = localtime(&runtimer);
 
     strftime(timestr, TIMESIZE, TIMEFORMAT_POSIX, runtime);
@@ -707,36 +748,6 @@ process_jobs(int argc, char **argv, int what)
     return rc;
 }				/* delete_jobs */
 
-static time_t parsetimespec(const char *spec)
-{
-    int len = strlen(spec);
-    struct tm exectm;
-    const char *dot;
-    const char *pos = spec;
-
-    exectm = *localtime( time( NULL ) );
-
-    dot = strrchr(spec, '.');
-    if (dot) {
-	len -= 3;
-    }
-
-    if (len == 12) {
-	pos += 4;
-	len -= 4;
-	/* first 4 digits to year */
-    } else if (len == 10) {
-	pos += 2;
-	len -= 2;
-	/* first 2 digits become last 2 digits of year.  If in past, +100 to year */
-    }
-    if (len == 8)
-	/* year is current year unless date would be in past, if so, next year */
-	;
-    else
-	return 0; /* corrupt */
-}
-
 /* Global functions */
 
 void *
@@ -754,14 +765,14 @@ int
 main(int argc, char **argv)
 {
     int c;
-    char queue = DEFAULT_AT_QUEUE;
+    char queue = DEFAULT_QUEUE;
     char queue_set = 0;
     char *pgm;
 
     int program = AT;		/* our default program */
-    char *options = "q:f:MmvldhVc";	/* default options for at */
+    char *options = "q:f:Mmu:bvlrdhVct:";	/* default options for at */
     int disp_version = 0;
-    time_t timer;
+    time_t timer = 0;
     struct passwd *pwe;
     struct group *ge;
 
@@ -816,6 +827,10 @@ main(int argc, char **argv)
 	    send_mail = -1;
 	    break;
 
+	case 'u':               /* send mail to specific user */
+	    mail_rcpt = optarg;
+	    break;
+
 	case 'f':
 	    atinput = optarg;
 	    break;
@@ -865,6 +880,13 @@ main(int argc, char **argv)
 	    options = "";
 	    break;
 
+	case 't':
+	    if (!posixtime(&timer, optarg, PDS_LEADING_YEAR | PDS_CENTURY | PDS_SECONDS)) {
+		fprintf(stderr, "invalid date format: %s\n", optarg);
+		exit(EXIT_FAILURE);
+	    }
+	    break;
+
 	default:
 	    usage();
 	    break;
@@ -872,9 +894,12 @@ main(int argc, char **argv)
     /* end of options eating
      */
 
-    if (disp_version)
+    if (disp_version) {
 	fprintf(stderr, "at version " VERSION "\n"
-	   "Bug reports to: rmurray@debian.org (Ryan Murray)\n");
+	   "Please report bugs to the Debian bug tracking system (http://bugs.debian.org/)\n"
+	   "or contact the maintainers (at@packages.debian.org).\n");
+	exit(EXIT_SUCCESS);
+    }
 
     /* select our program
      */
@@ -922,9 +947,11 @@ main(int argc, char **argv)
 
     case AT:
 	if (argc > optind) {
-	    timer = parsetime(argc - optind, argv + optind);
-	} else {
-	    timer = 0;
+	    if (timer != 0) {
+                fprintf(stderr, "Cannot give time twice.\n");
+                exit(EXIT_FAILURE);
+            }
+	    timer = parsetime(time(0), argc - optind, argv + optind);
 	}
 
 	if (timer == 0) {
@@ -951,11 +978,15 @@ main(int argc, char **argv)
 	if (queue_set)
 	    queue = toupper(queue);
 	else
-	    queue = DEFAULT_BATCH_QUEUE;
+	    queue = BATCH_QUEUE;
 
-	if (argc > optind)
-	    timer = parsetime(argc, argv);
-	else
+	if (argc > optind) {
+            if (timer != 0) {
+                fprintf(stderr, "Cannot give time twice.\n");
+                exit(EXIT_FAILURE);
+            }
+	    timer = parsetime(time(0), argc, argv);
+        } else if (timer == 0)
 	    timer = time(NULL);
 
 	if (atverify) {
